#  !/usr/bin/env python
#
#  templates.py
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
# this package
from .templates import templates

sectioning_template = templates.get_template("sectioning.tex")

# 			body=pathlib.Path(filename).read_text(),


def _make_section(section_type, title, body='', label=None, shorttitle=None, unnumbered=False):

	if not label:
		label = f"{section_type}:{title.lower().replace(' ', '_')}"

	return sectioning_template.render(
			section_type=section_type,
			title=title,
			label=label,
			body=body,
			shorttitle=shorttitle,
			unnumbered=unnumbered,
			)


def make_part(title, body='', label=None, shorttitle=None, unnumbered=False):
	return _make_section(
			"part",
			title=title,
			label=label,
			body=body,
			shorttitle=shorttitle,
			unnumbered=unnumbered,
			)


def make_chapter(title, body='', label=None, shorttitle=None, unnumbered=False):
	return _make_section(
			"chapter",
			title=title,
			label=label,
			body=body,
			shorttitle=shorttitle,
			unnumbered=unnumbered,
			)


def make_section(title, body='', label=None, shorttitle=None, unnumbered=False):
	return _make_section(
			"section",
			title=title,
			label=label,
			body=body,
			shorttitle=shorttitle,
			unnumbered=unnumbered,
			)


def make_subsection(title, body='', label=None, shorttitle=None, unnumbered=False):
	return _make_section(
			"subsection",
			title=title,
			label=label,
			body=body,
			shorttitle=shorttitle,
			unnumbered=unnumbered,
			)


def make_subsubsection(title, body='', label=None, shorttitle=None, unnumbered=False):
	return _make_section(
			"subsubsection",
			title=title,
			label=label,
			body=body,
			shorttitle=shorttitle,
			unnumbered=unnumbered,
			)


def make_paragraph(title, body='', label=None, shorttitle=None, unnumbered=False):
	return _make_section(
			"paragraph",
			title=title,
			label=label,
			body=body,
			shorttitle=shorttitle,
			unnumbered=unnumbered,
			)


def make_subparagraph(title, body='', label=None, shorttitle=None, unnumbered=False):
	return _make_section(
			"subparagraph",
			title=title,
			label=label,
			body=body,
			shorttitle=shorttitle,
			unnumbered=unnumbered,
			)



#!/usr/bin/env python3
#
#  __init__.py
"""
Docstring Goes Here
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

# stdlib
import os
import pathlib
from typing import Iterable, Optional, Union

# this package
from .templates import templates

__author__ = "Dominic Davis-Foster"
__copyright__ = "2020 Dominic Davis-Foster"

__license__ = "GNU Lesser General Public License v3 or later (LGPLv3+)"
__version__ = "0.0.4"
__email__ = "dominic@davis-foster.co.uk"

main_template = templates.get_template("main.tex")


def make_document(
		outfile: Union[str, pathlib.Path, os.PathLike],
		elements: Optional[Iterable[str]] = None,
		glossary: str = '',
		):

	if not isinstance(outfile, pathlib.Path):
		outfile = pathlib.Path(outfile)

	if elements is None:
		elements = []

	with open(outfile, "w") as fp:
		fp.write(main_template.render(elements=elements, glossary=glossary))



#  !/usr/bin/env python
#
#  templates.py
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import pathlib

# 3rd party
from jinja2 import Environment, FileSystemLoader

__all__ = ["template_dir", "templates"]

template_dir = (pathlib.Path(__file__).parent / "templates").absolute()

templates = Environment(loader=FileSystemLoader(template_dir))  # type: ignore



#  !/usr/bin/env python
#
#  tables.py
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import re
from functools import partial
from textwrap import indent
from typing import Iterable, List, Optional, Sequence, Union

# 3rd party
import tabulate
from tabulate import Line, TableFormat

# this package
from py2latex.core import begin, make_caption, make_label, re_escape
from py2latex.templates import templates

hline = r"\hline"
toprule = r"\toprule"
midrule = r"\midrule"
endfoot = r"\endfoot"
endhead = r"\endhead"
bottomrule = r"\bottomrule"
endlastfoot = r"\endlastfoot"
table_linebreak = r"\\"


def multicolumn(cols: int, pos: str, text: str) -> str:
	"""

	:param cols: The number of columms to span
	:type cols: int
	:param pos: Text alignment:
		* c for centered
		* l for flushleft
		* r for flushright
	:type pos: str
	:param text:
	:type text: str

	:return:
	:rtype: str
	"""

	return fr"\multicolumn{{{cols}}}{{{pos}}}{{{{{text}}}}}"


def _latex_line_begin_tabular(colwidths, colaligns, booktabs=False, longtable=False, longtable_continued=False):
	"""
	Based on Bart Broere's fork of python-tabulate
	https://github.com/bartbroere/python-tabulate
	MIT Licensed

	https://pypi.org/project/tabulate/
	"""

	alignment = {"left": "l", "right": "r", "center": "c", "decimal": "r"}
	tabular_columns_fmt = "".join([alignment.get(a, "l") for a in colaligns])

	if longtable:

		line_begin_elements = [rf"\begin{{longtable}}{{{tabular_columns_fmt}}}"]

		if longtable_continued:
			line_begin_elements.append(
					fr"""\midrule
\multicolumn{{{len(colwidths)}}}{{r}}{{{{\small Continued on next page\normalsize}}}} \\
\midrule
\endfoot

\bottomrule
\endlastfoot
"""
					)
	else:
		line_begin_elements = [rf"\begin{{tabular}}{{{tabular_columns_fmt}}}"]

	if booktabs:
		line_begin_elements.append(toprule)
	else:
		line_begin_elements.append(hline)

	return "\n".join(line_begin_elements)


def latex_format_builder(
		raw: bool = False,
		booktabs: bool = False,
		longtable: bool = False,
		longtable_continued: bool = False,
		) -> TableFormat:
	lineabove = partial(
			_latex_line_begin_tabular,
			longtable=longtable,
			booktabs=booktabs,
			longtable_continued=longtable_continued
			)

	if booktabs and longtable:
		linebelowheader = [midrule, endhead]

		if longtable_continued:
			linebelow = [r"\end{longtable}"]
		else:
			linebelow = [bottomrule, r"\end{longtable}"]

	elif booktabs:
		linebelowheader = [midrule]
		linebelow = [bottomrule, r"\end{tabular}"]

	elif longtable:
		linebelowheader = [hline, endhead]
		linebelow = [hline, r"\end{longtable}"]

	else:
		linebelowheader = [hline]
		linebelow = [hline, r"\end{tabular}"]

	if raw:
		datarow = headerrow = partial(tabulate._latex_row, escrules={})  # type: ignore
	else:
		datarow = headerrow = tabulate._latex_row  # type: ignore

	return TableFormat(
			lineabove=lineabove,
			linebelowheader=Line("\n".join(linebelowheader), "", "", ""),
			linebelow=Line("\n".join(linebelow), "", "", ""),
			datarow=datarow,
			headerrow=headerrow,
			linebetweenrows=None,
			padding=1,
			with_header_hide=None,
			)


table_formats = {
		"latex": latex_format_builder(),
		"latex_raw": latex_format_builder(raw=True),
		"latex_booktabs": latex_format_builder(booktabs=True),
		"latex_booktabs_raw": latex_format_builder(raw=True, booktabs=True),
		"latex_longtable": latex_format_builder(longtable=True),
		"latex_longtable_raw": latex_format_builder(raw=True, longtable=True),
		"latex_longtable_booktabs": latex_format_builder(longtable=True, booktabs=True),
		"latex_longtable_booktabs_raw": latex_format_builder(raw=True, longtable=True, booktabs=True),
		}

raw_longbooktab_cont = latex_format_builder(raw=True, booktabs=True, longtable=True, longtable_continued=True)


def add_longtable_caption(table: str, caption: Optional[str] = None, label: Optional[str] = None) -> str:
	"""
	Add a caption to a longtable

	:param table:
	:type table: str
	:param caption: str
	:type caption:
	:param label:
	:type label: str

	:return:
	:rtype:
	"""

	elems = []

	if caption:
		elems.append(make_caption(caption))

	if label:
		elems.append(make_label(label))

	if caption or label:
		table = table.replace(toprule, "".join([*elems, "\\\\\n", toprule]))
	# table = table.replace(
	# 		toprule,
	# 		"".join([*elems, "\\\\\n", toprule, "\n\\endfirsthead\n", f"\\caption*{{{caption}}}\\\\\n"])
	# 		)

	return table


def set_table_widths(table: str, widths: str) -> str:
	"""
	Override the column widths (and also the column alignments) in a tabular environment, etc.

	:param table:
	:type table: str
	:param widths:
	:type widths: str

	:return:
	:rtype: str
	"""

	for environment in {"tabular", "longtable"}:
		table = re.sub(
				fr"{re_escape(begin(environment))}(\[[^\]]*\])?{{.*}}",
				fr"{re_escape(begin(environment))}{{{widths}}}",
				table,
				)

	return table


_longtable_template = templates.get_template("longtable.tex")
_table_template = templates.get_template("table.tex")
_subtables_template = templates.get_template("subtables.tex")
_subtables_template.globals["indent"] = indent  # type: ignore


def longtable_from_template(
		tabular_data,
		*,
		caption: str,
		label: Optional[str] = None,
		headers: Sequence[str] = (),
		pos: str = "htpb",
		floatfmt: str = tabulate._DEFAULT_FLOATFMT,  # type: ignore
		numalign: str = "decimal",
		stralign: str = "left",
		missingval: str = tabulate._DEFAULT_MISSINGVAL,  # type: ignore
		showindex: str = "default",
		disable_numparse: bool = False,
		colalign: Optional[Sequence[Union[str, None]]] = None,
		colwidths: Optional[Sequence[Union[str, None]]] = None,
		vlines: Union[Sequence[int], bool] = False,
		hspace: Union[Sequence[int], bool] = False,
		raw: bool = True,
		footer: Optional[str] = None,
		) -> str:
	"""
	Create a ``longtable`` with ``booktabs`` formatting.

	:param tabular_data:
	:type tabular_data:
	:param caption: The caption for the table
	:type caption: str
	:param label: The label for the table.
		If undefined the caption is used, in lowercase, with underscores replacing spaces
	:param headers: A sequence of column headers
	:param pos: The positioning of the table, e.g. ``"htp"``
	:type pos: str
	:param floatfmt: The formatting of :class:`float` values. Default ``"g"``
	:type floatfmt: str
	:param numalign:
	:type numalign:
	:param stralign:
	:type stralign:
	:param missingval:
	:type missingval:
	:param showindex:
	:type showindex:
	:param disable_numparse:
	:type disable_numparse:
	:param colalign:
	:param colwidths: Sequence of column widths, e.g. ``3cm``. Values of :py:obj:`None` indicates auto width
	:param vlines: If a sequence of integers a line will be inserted before the specified columns. ``-1`` indicates a line should be inserted after the last column.
		If :py:obj:`True` a line will be inserted before every column, and after the last column.
		If :py:obj:`False` no lines will be inserted.
	:param hspace: If a sequence of integers extra space will be inserted before the specified row. ``-1`` indicates a space should be inserted after the last row.
		If :py:obj:`True` a space will be inserted before every row, and after the last row.
		If :py:obj:`False` no spaces will be inserted.
	:param raw: Whether latex markup in ``tabular_data`` should be unescaped. Default ``True``
	:type raw: bool
	:param footer: Optional footer for the table. Inserted as raw LaTeX

	:return:
	:rtype: str
	"""

	return table_from_template(
			tabular_data,
			caption=caption,
			label=label,
			headers=headers,
			pos=pos,
			floatfmt=floatfmt,
			numalign=numalign,
			stralign=stralign,
			missingval=missingval,
			showindex=showindex,
			disable_numparse=disable_numparse,
			colalign=colalign,
			colwidths=colwidths,
			vlines=vlines,
			hspace=hspace,
			raw=raw,
			footer=footer,
			longtable=True,
			)


def _make_body_only_formats(raw=False):

	if raw:
		datarow = headerrow = partial(tabulate._latex_row, escrules={})  # type: ignore
	else:
		datarow = headerrow = tabulate._latex_row  # type: ignore

	return TableFormat(
			lineabove=None,
			linebelowheader=None,
			linebelow=None,
			datarow=datarow,
			headerrow=headerrow,
			linebetweenrows=None,
			padding=1,
			with_header_hide=None,
			)


body_only_format = _make_body_only_formats(raw=False)
raw_body_only_format = _make_body_only_formats(raw=True)


def _parse_rows(
		rows: List[str],
		tabular_data,
		headers: Sequence[str] = (),
		):

	if headers:
		header_row = rows[0]
		body_rows = rows[1:]
		ncols = len(headers)
	else:
		header_row = ''
		body_rows = rows
		ncols = len(tabular_data[0])

	return header_row, body_rows, ncols


def parse_hspace(ncols: int, hspace: Union[Sequence[int], bool] = False):

	if isinstance(hspace, Sequence):
		add_hspace = True
	else:
		add_hspace = bool(hspace)
		hspace = list(range(ncols))

	return add_hspace, hspace


def table_from_template(
		tabular_data,
		*,
		caption: str,
		label: Optional[str] = None,
		headers: Sequence[str] = (),
		pos: str = "htpb",
		floatfmt: str = tabulate._DEFAULT_FLOATFMT,  # type: ignore
		numalign: str = "decimal",
		stralign: str = "left",
		missingval: str = tabulate._DEFAULT_MISSINGVAL,  # type: ignore
		showindex: str = "default",
		disable_numparse: bool = False,
		colalign: Optional[Sequence[Union[str, None]]] = None,
		colwidths: Optional[Sequence[Union[str, None]]] = None,
		vlines: Union[Sequence[int], bool] = False,
		hspace: Union[Sequence[int], bool] = False,
		raw: bool = True,
		footer: Optional[str] = None,
		longtable: bool = False
		) -> str:
	"""
	Create a ``table`` with ``booktabs`` formatting.

	:param tabular_data:
	:type tabular_data:
	:param caption: The caption for the table
	:type caption: str
	:param label: The label for the table.
		If undefined the caption is used, in lowercase, with underscores replacing spaces
	:param headers: A sequence of column headers
	:param pos: The positioning of the table, e.g. ``"htp"``
	:type pos: str
	:param floatfmt: The formatting of :class:`float` values. Default ``"g"``
	:type floatfmt: str
	:param numalign:
	:type numalign:
	:param stralign:
	:type stralign:
	:param missingval:
	:type missingval:
	:param showindex:
	:type showindex:
	:param disable_numparse:
	:type disable_numparse:
	:param colalign:
	:param colwidths: Sequence of column widths, e.g. ``3cm``. Values of :py:obj:`None` indicates auto width
	:param vlines: If a sequence of integers a line will be inserted before the specified columns. ``-1`` indicates a line should be inserted after the last column.
		If :py:obj:`True` a line will be inserted before every column, and after the last column.
		If :py:obj:`False` no lines will be inserted.
	:param hspace: If a sequence of integers extra space will be inserted before the specified row. ``-1`` indicates a space should be inserted after the last row.
		If :py:obj:`True` a space will be inserted before every row, and after the last row.
		If :py:obj:`False` no spaces will be inserted.
	:param raw: Whether latex markup in ``tabular_data`` should be unescaped. Default ``False``
	:type raw: bool
	:param footer: Optional footer for the table. Inserted as raw LaTeX
	:param longtable: Whether to create a ``longtable``. Default :py:obj:`False``
	:type longtable: bool

	:return:
	:rtype: str
	"""

	table = SubTable(
			tabular_data,
			caption=caption,
			label=label,
			headers=headers,
			floatfmt=floatfmt,
			numalign=numalign,
			stralign=stralign,
			missingval=missingval,
			showindex=showindex,
			disable_numparse=disable_numparse,
			colalign=colalign,
			colwidths=colwidths,
			vlines=vlines,
			hspace=hspace,
			raw=raw,
			footer=footer,
			)

	if longtable:
		return _longtable_template.render(
				caption=table.caption,
				label=table.label,
				header_row=table.header_row,
				table_body=table.table_body,
				ncols=table.ncols,
				colalign=table.colalign,
				pos=pos,
				footer=table.footer,
				)

	else:
		return _table_template.render(
				caption=table.caption,
				label=table.label,
				header_row=table.header_row,
				table_body=table.table_body,
				ncols=table.ncols,
				colalign=table.colalign,
				pos=pos,
				footer=table.footer,
				)


class SubTable:

	def __init__(
			self,
			tabular_data,
			*,
			caption: str,
			label: Optional[str] = None,
			headers: Sequence[str] = (),
			floatfmt: str = tabulate._DEFAULT_FLOATFMT,  # type: ignore
			numalign: str = "decimal",
			stralign: str = "left",
			missingval: str = tabulate._DEFAULT_MISSINGVAL,  # type: ignore
			showindex: str = "default",
			disable_numparse: bool = False,
			colalign: Optional[Sequence[Union[str, None]]] = None,
			colwidths: Optional[Sequence[Union[str, None]]] = None,
			vlines: Union[Sequence[int], bool] = False,
			hspace: Union[Sequence[int], bool] = False,
			raw: bool = True,
			footer: Optional[str] = None,
			) -> None:

		if raw:
			tablefmt = raw_body_only_format
		else:
			tablefmt = body_only_format

		rows = tabulate.tabulate(
				tabular_data,
				tablefmt=tablefmt,
				headers=headers,
				floatfmt=floatfmt,
				numalign=numalign,
				stralign=stralign,
				missingval=missingval,
				showindex=showindex,
				disable_numparse=disable_numparse,  # colalign=colalign,
				).split("\n")

		header_row, body_rows, ncols = _parse_rows(rows, tabular_data, headers)
		add_hspace, hspace = parse_hspace(ncols, hspace)
		col_alignment = parse_column_alignments(colalign, colwidths, vlines, ncols)

		table_body = ''
		for row_idx, row in enumerate(body_rows):
			row = re.sub(r"(\\multicolumn{2\}{.*\}{{.*\}\}\s*)&", r"\1 ", row)
			row = re.sub(r"(\\multicolumn{3\}{.*\}{{.*\}\}\s*)&(\s*)&", r"\1 \2", row)
			row = re.sub(r"(\\multicolumn{4\}{.*\}{{.*\}\}\s*)&(\s*)&(\s*)&", r"\1 \2 \3", row)
			row = re.sub(r"(\\multicolumn{5\}{.*\}{{.*\}\}\s*)&(\s*)&(\s*)&(\s*)&", r"\1 \2 \3 \4", row)

			if add_hspace and row_idx in hspace:  # type: ignore
				table_body += "\\addlinespace"

			table_body += f"{row}\n"

		if not label:
			label = caption.lower().replace(" ", "_")

		self.caption: str = str(caption)
		self.label: str = str(label)
		self.header_row: str = header_row
		self.table_body: str = indent(table_body, "       ")
		self.ncols: int = ncols
		self.colalign: str = "".join(col_alignment)
		self.footer: str = str(footer)


def subtables_from_template(
		subtables: Iterable[SubTable],
		*,
		caption: str,
		label: Optional[str] = None,
		pos: str = "htpb",
		) -> str:
	# TODO: customise spacing between tables
	"""
	Create a series of ``subtables`` with ``booktabs`` formatting.

	:param subtables:
	:type subtables:
	:param caption: The caption for the table
	:type caption: str
	:param label: The label for the table.
		If undefined the caption is used, in lowercase, with underscores replacing spaces
	:param pos: The positioning of the table, e.g. ``"htp"``
	:type pos: str

	:return:
	:rtype: str
	"""

	if not label:
		label = caption.lower().replace(" ", "_")

	return _subtables_template.render(
			subtables=subtables,
			caption=caption,
			label=label,
			pos=pos,
			)


def parse_column_alignments(colalign, colwidths, vlines, ncols):
	if colalign is None:
		colalign = ["l"] * ncols

	while len(colalign) < ncols:
		colalign.append("l")

	alignment_elements = []

	if colwidths is None:
		colwidths = [None] * ncols

	while len(colwidths) < ncols:
		colwidths.append(None)

	if isinstance(vlines, Sequence):
		add_vlines = True
	else:
		add_vlines = bool(vlines)
		vlines = list(range(ncols + 1))

	col_idx = 0

	for col_idx, (alignment, width) in enumerate(zip(colalign, colwidths)):
		# print(col_idx, alignment, width)

		if add_vlines and col_idx in vlines:
			alignment_elements.append("|")

		if alignment.startswith("l"):
			if width is None:
				alignment_elements.append("l")
			else:
				alignment_elements.append(fr">{{\raggedright}}p{{{width}}}")

		elif alignment.startswith("r") or alignment.lower() == "decimal":
			if width is None:
				alignment_elements.append("r")
			else:
				alignment_elements.append(fr">{{\raggedleft\arraybackslash}}p{{{width}}}")

		# TODO: centered and width

		elif alignment.startswith("c"):
			alignment_elements.append("c")

		elif alignment.startswith("p"):
			if width is None:
				raise ValueError(f"Must specify width for 'p' column with index {col_idx}")
			else:
				alignment_elements.append(fr">{{\raggedleft\arraybackslash}}p{{{width}}}")

	if add_vlines:
		if col_idx + 1 in vlines or -1 in vlines:
			alignment_elements.append("|")

	return "".join(alignment_elements)


templates.globals["brace"] = lambda var: f"{{{var}}}"

if __name__ == '__main__':

	# stdlib
	from pprint import pprint

	pprint(
			longtable_from_template([
					[1, 2, 3, 4, 5],
					[1, 2, 3, 4, 5],
					[1, 2, 3, 4, 5],
					[1, 2, 3, 4, 5],
					],
									headers=["Foo", "Bar", "Baz", "Fizz", "Buzz"],
									caption="My Caption")
			)



# From https://github.com/rufuspollock/markdown2latex
# BSD Licensed

#!/usr/bin/env python3
"""Extension to python-markdown to support LaTeX (rather than html) output.

Authored by Rufus Pollock: <http://www.rufuspollock.org/>
Reworked by Julian Wulfheide (ju.wulfheide@gmail.com) and
Pedro Gaudencio (pmgaudencio@gmail.com)

Usage:
======

1. Command Line. A script entitled markdown2latex.py is automatically
installed. For details of usage see help::

	$ markdown2latex.py -h

2. As a python-markdown extension::

	>>> import markdown
	>>> md = markdown.Markdown(None, extensions=['latex'])
	>>> # text is input string ...
	>>> latex_out = md.convert(text)

3. Directly as a module (slight inversion of std markdown extension setup)::

	>>> import markdown
	>>> import mdx_latex
	>>> md = markdown.Markdown()
	>>> latex_mdx = mdx_latex.LaTeXExtension()
	>>> latex_mdx.extendMarkdown(md, markdown.__dict__)
	>>> out = md.convert(text)

History
=======

Version: 1.0 (November 15, 2006)

  * First working version (compatible with markdown 1.5)
  * Includes support for tables

Version: 1.1 (January 17, 2007)

  * Support for verbatim and images

Version: 1.2 (June 2008)

  * Refactor as an extension.
  * Make into a proper python/setuptools package.
  * Tested with markdown 1.7 but should work with 1.6 and (possibly) 1.5
	(though pre/post processor stuff not as worked out there)

Version 1.3: (July 2008)
  * Improvements to image output (width)

Version 1.3.1: (August 2009)
  * Tiny bugfix to remove duplicate keyword argument and set zip_safe=False
  * Add [width=\textwidth] by default for included images

Version 2.0: (June 2011)
  * PEP8 cleanup
  * Major rework since this was broken by new Python-Markdown releases

Version 2.1: (August 2013)
  * Add handler for non locally referenced images, hyperlinks and horizontal rules
  * Update math delimiters
"""

__version__ = '2.1'

# stdlib
import http.client
import os
# do some fancy importing stuff to allow use to override things in this module
# in this file while still importing * for use in our own classes
import re
import sys
import tempfile
import urllib.error
import urllib.parse
import urllib.request
import xml.dom.minidom
from urllib.parse import urlparse

# 3rd party
import markdown  # type: ignore

start_single_quote_re = re.compile("(^|\\s|\")'")
start_double_quote_re = re.compile("(^|\\s|'|`)\"")
end_double_quote_re = re.compile("\"(,|\\.|\\s|$)")


def unescape_html_entities(text):
	out = text.replace('&amp;', '&')
	out = out.replace('&lt;', '<')
	out = out.replace('&gt;', '>')
	out = out.replace('&quot;', '"')
	return out


def escape_latex_entities(text):
	"""Escape latex reserved characters."""
	out = text
	out = unescape_html_entities(out)

	out = re.sub(r"[^\n\\]%", r"\\%", out)
	out = re.sub(r"[^\\]&", r"\\&", out)
	out = re.sub(r"[^\\]#", r"\\#", out)

	out = re.sub(r'\"([^\"]*)\"', r'\\enquote{\1}', out)
	out = re.sub(r"\'([^\']*)\'", r'\\enquote{\1}', out)

	# out = start_single_quote_re.sub(r'\g<1>`', out)
	# out = start_double_quote_re.sub(r'\g<1>``', out)
	# out = end_double_quote_re.sub(r"''\g<1>", out)

	# people should escape these themselves as it conflicts with maths
	# out = out.replace('{', '\\{')
	# out = out.replace('}', '\\}')
	# do not do '$' here because it is dealt with by convert_maths
	# out = out.replace('$', '\\$')
	return out


def unescape_latex_entities(text):
	"""Limit ourselves as this is only used for maths stuff."""
	out = text
	out = out.replace('\\&', '&')
	return out


def makeExtension(configs=None):
	return LaTeXExtension(configs=configs)


class LaTeXExtension(markdown.Extension):

	def __init__(self, configs=None):
		self.reset()

	def extendMarkdown(self, md, md_globals):
		self.md = md

		# remove escape pattern -- \\(.*) -- as this messes up any embedded
		# math and we don't need to escape stuff any more for html
		# for key, pat in self.md.inlinePatterns.items():
		#     if pat.pattern == markdown.inlinepatterns.ESCAPE_RE:
		#         self.md.inlinePatterns.pop(key)
		#         break

		#footnote_extension = FootnoteExtension()
		#footnote_extension.extendMarkdown(md, md_globals)

		latex_tp = LaTeXTreeProcessor()
		math_pp = MathTextPostProcessor()
		table_pp = TableTextPostProcessor()
		image_pp = ImageTextPostProcessor()
		link_pp = LinkTextPostProcessor()
		unescape_html_pp = UnescapeHtmlTextPostProcessor()

		md.treeprocessors['latex'] = latex_tp
		md.postprocessors['unescape_html'] = unescape_html_pp
		md.postprocessors['math'] = math_pp
		md.postprocessors['image'] = image_pp
		md.postprocessors['table'] = table_pp
		md.postprocessors['link'] = link_pp

	def reset(self):
		pass


class LaTeXTreeProcessor(markdown.treeprocessors.Treeprocessor):

	def run(self, doc):
		"""Walk the dom converting relevant nodes to text nodes with relevant
		content."""
		latex_text = self.tolatex(doc)

		doc.clear()
		latex_node = markdown.util.etree.Element('root')
		latex_node.text = latex_text
		doc.append(latex_node)

	def tolatex(self, ournode):
		buffer = ""
		subcontent = ""

		if ournode.text:
			subcontent += escape_latex_entities(ournode.text)

		if ournode.getchildren():
			for child in ournode.getchildren():
				subcontent += self.tolatex(child)

		if ournode.tag == 'h1':
			buffer += f'\n\\chapter{{{subcontent}}}'
			buffer += f'\n\\label{{chapter:{subcontent.lower().replace(" ", "_")}}}\n'

		elif ournode.tag == 'h2':
			buffer += f'\n\n\\section{{{subcontent}}}'
			buffer += f'\n\\label{{section:{subcontent.lower().replace(" ", "_")}}}\n'
		elif ournode.tag == 'h3':
			buffer += f'\n\n\\subsection{{{subcontent}}}'
			buffer += f'\n\\label{{subsection:{subcontent.lower().replace(" ", "_")}}}\n'
		elif ournode.tag == 'h4':
			buffer += f'\n\\subsubsection{{{subcontent}}}'
			buffer += f'\n\\label{{subsubsection:{subcontent.lower().replace(" ", "_")}}}\n'
		elif ournode.tag == 'hr':
			buffer += '\\noindent\\makebox[\\linewidth]{\\rule{\\linewidth}{0.4pt}}'
		elif ournode.tag == 'ul':
			# no need for leading \n as one will be provided by li
			buffer += f"""
\\begin{{itemize}}{subcontent}
\\end{{itemize}}
"""
		elif ournode.tag == 'ol':
			# no need for leading \n as one will be provided by li
			buffer += f"""
\\begin{{enumerate}}{subcontent}
\\end{{enumerate}}
"""
		elif ournode.tag == 'li':
			buffer += f"""
  \\item {subcontent.strip()}"""
		elif ournode.tag == 'blockquote':
			# use quotation rather than quote as quotation can support multiple
			# paragraphs
			buffer += f"""
\\begin{{quotation}}
{subcontent.strip()}
\\end{{quotation}}
"""
		# ignore 'code' when inside pre tags
		# (mkdn produces <pre><code></code></pre>)
		elif (
				ournode.tag == 'pre' or
				# TODO: Take a look here
				(ournode.tag == 'pre' and ournode.parentNode.tag != 'pre')
				):
			buffer += f"""
\\begin{{verbatim}}
{subcontent.strip()}
\\end{{verbatim}}
"""
		elif ournode.tag == 'q':
			buffer += f"`{subcontent.strip()}'"
		elif ournode.tag == 'p':
			buffer += f'\n{subcontent.strip()}\n'
		# Footnote processor inserts all of the footnote in a sup tag
		elif ournode.tag == 'sup':
			buffer += f"\\footnote{{{subcontent.strip()}}}"
		elif ournode.tag == 'strong':
			buffer += f'\\textbf{{{subcontent.strip()}}}'
		elif ournode.tag == 'em':
			buffer += f"\\emph{{{subcontent.strip()}}}"
		# Keep table strcuture. TableTextPostProcessor will take care.
		elif ournode.tag == 'table':
			buffer += f"\n\n<table>{subcontent}</table>\n\n"
		elif ournode.tag == 'thead':
			buffer += f"<thead>{subcontent}</thead>"
		elif ournode.tag == 'tbody':
			buffer += f"<tbody>{subcontent}</tbody>"
		elif ournode.tag == 'tr':
			buffer += f"<tr>{subcontent}</tr>"
		elif ournode.tag == 'th':
			buffer += f"<th>{subcontent}</th>"
		elif ournode.tag == 'td':
			buffer += f"<td>{subcontent}</td>"
		elif ournode.tag == 'img':
			buffer += f'<img src=\"{ournode.get("src")}\" alt=\"{ournode.get("alt")}\" />'
		elif ournode.tag == 'a':
			buffer += f'<a href=\"{ournode.get("href")}\">{subcontent}</a>'
		else:
			buffer = subcontent

		if ournode.tail:
			buffer += escape_latex_entities(ournode.tail)

		return buffer


class UnescapeHtmlTextPostProcessor(markdown.postprocessors.Postprocessor):

	def run(self, text):
		return unescape_html_entities(text)


# ========================= MATHS =================================


class MathTextPostProcessor(markdown.postprocessors.Postprocessor):

	def run(self, instr):
		"""Convert all math sections in {text} whether latex, asciimathml or
		latexmathml formatted to latex.

		This assumes you are using $ for inline math and $$ for blocks as your
		mathematics delimiter (*not* the standard asciimathml or latexmathml
		delimiter).
		"""

		def repl_1(matchobj):
			text = unescape_latex_entities(matchobj.group(1))
			return f"\\[{text}\\]"

		def repl_2(matchobj):
			text = unescape_latex_entities(matchobj.group(1))
			return f"\\({text}\\)"

		# This $$x=3$$ is block math
		pat = re.compile(r'\$\$([^\$]*)\$\$')
		out = pat.sub(repl_1, instr)
		# This $x=3$ is inline math
		pat2 = re.compile(r'\$([^\$]*)\$')
		out = pat2.sub(repl_2, out)
		# some extras due to asciimathml
		out = out.replace('\\lt', '<')
		out = out.replace(' * ', ' \\cdot ')
		out = out.replace('\\del', '\\partial')
		return out


# ========================= TABLES =================================


class TableTextPostProcessor(markdown.postprocessors.Postprocessor):

	def run(self, instr):
		"""This is not very sophisticated and for it to work it is expected
		that:
			1. tables to be in a section on their own (that is at least one
			blank line above and below)
			2. no nesting of tables
		"""
		converter = Table2Latex()
		new_blocks = []

		for block in instr.split('\n\n'):
			stripped = block.strip()
			# <table catches modified verions (e.g. <table class="..">
			if stripped.startswith('<table') and stripped.endswith('</table>'):
				latex_table = converter.convert(stripped).strip()
				new_blocks.append(latex_table)
			else:
				new_blocks.append(block)
		return '\n\n'.join(new_blocks)


class Table2Latex:
	"""
	Convert html tables to Latex.

	TODO: escape latex entities.
	"""

	def colformat(self):
		# centre align everything by default
		out = '|l' * self.maxcols + '|'
		return out

	def get_text(self, element):
		if element.nodeType == element.TEXT_NODE:
			return escape_latex_entities(element.data)
		result = ''
		if element.childNodes:
			for child in element.childNodes:
				text = self.get_text(child)
				if text.strip() != '':
					result += text
		return result

	def process_cell(self, element):
		# works on both td and th
		colspan = 1
		subcontent = self.get_text(element)
		buffer = ""

		if element.tagName == 'th':
			subcontent = f"\\textbf{{{subcontent}}}"
		if element.hasAttribute('colspan'):
			colspan = int(element.getAttribute('colspan'))
			buffer += fr' \multicolumn{{{colspan}}}{{|c|}}{{{subcontent}}}'
		# we don't support rowspan because:
		#   1. it needs an extra latex package \usepackage{multirow}
		#   2. it requires us to mess around with the alignment tags in
		#   subsequent rows (i.e. suppose the first col in row A is rowspan 2
		#   then in row B in the latex we will need a leading &)
		# if element.hasAttribute('rowspan'):
		#     rowspan = int(element.getAttribute('rowspan'))
		#     buffer += ' \multirow{%s}{|c|}{%s}' % (rowspan, subcontent)
		else:
			buffer += f" {subcontent}"

		notLast = (
				element.nextSibling.nextSibling
				and element.nextSibling.nextSibling.nodeType == element.ELEMENT_NODE
				and element.nextSibling.nextSibling.tagName in ['td', 'th']
				)

		if notLast:
			buffer += ' &'

		self.numcols += colspan
		return buffer

	def tolatex(self, element):
		if element.nodeType == element.TEXT_NODE:
			return ""

		buffer = ""
		subcontent = ""
		if element.childNodes:
			for child in element.childNodes:
				text = self.tolatex(child)
				if text.strip() != "":
					subcontent += text
		subcontent = subcontent.strip()

		if element.tagName == 'thead':
			buffer += subcontent

		elif element.tagName == 'tr':
			self.maxcols = max(self.numcols, self.maxcols)
			self.numcols = 0
			buffer += f"\n\\hline\n{subcontent} \\\\"

		elif element.tagName == 'td' or element.tagName == 'th':
			buffer = self.process_cell(element)
		else:
			buffer += subcontent
		return buffer

	def convert(self, instr):
		self.numcols = 0
		self.maxcols = 0
		dom = xml.dom.minidom.parseString(instr)
		core = self.tolatex(dom.documentElement)

		captionElements = dom.documentElement.getElementsByTagName('caption')
		caption = ''
		if captionElements:
			caption = self.get_text(captionElements[0])

		colformatting = self.colformat()
		table_latex = \
                     f"""
			\\begin{{table}}[h]
			\\begin{{tabular}}{{{colformatting}}}
			{core}
			\\hline
			\\end{{tabular}}
			\\\\[5pt]
			\\caption{{{caption}}}
			\\end{{table}}
			"""
		return table_latex


# ========================= IMAGES =================================


class ImageTextPostProcessor(markdown.postprocessors.Postprocessor):

	def run(self, instr):
		"""Process all img tags

		Similar to process_tables this is not very sophisticated and for it
		to work it is expected that img tags are put in a section of their own
		(that is separated by at least one blank line above and below).
		"""
		converter = Img2Latex()
		new_blocks = []
		for block in instr.split("\n\n"):
			stripped = block.strip()
			# <table catches modified verions (e.g. <table class="..">
			if stripped.startswith('<img'):
				latex_img = converter.convert(stripped).strip()
				new_blocks.append(latex_img)
			else:
				new_blocks.append(block)
		return '\n\n'.join(new_blocks)


class Img2Latex:

	def convert(self, instr):
		dom = xml.dom.minidom.parseString(instr)
		img = dom.documentElement
		src = img.getAttribute('src')

		if urlparse(src).scheme != '':
			src_urlparse = urlparse(src)
			conn = http.client.HTTPConnection(src_urlparse.netloc)
			conn.request('HEAD', src_urlparse.path)
			response = conn.getresponse()
			conn.close()
			if response.status == 200:
				filename = os.path.join(tempfile.mkdtemp(), src.split('/')[-1])
				urllib.request.urlretrieve(src, filename)
				src = filename

		alt = img.getAttribute('alt')
		# Using graphicx and ajustbox package for *max width*
		out = \
                     f"""
			\\begin{{figure}}[H]
			\\centering
			\\includegraphics[max width=\\linewidth]{{{src}}}
			\\caption{{{alt}}}
			\\end{{figure}}
			"""
		return out


# ========================== LINKS =================================


class LinkTextPostProcessor(markdown.postprocessors.Postprocessor):

	def run(self, instr):
		# Process all hyperlinks
		converter = Link2Latex()
		new_blocks = []
		for block in instr.split("\n\n"):
			stripped = block.strip()
			match = re.search(r'<a[^>]*>([^<]+)</a>', stripped)
			# <table catches modified verions (e.g. <table class="..">
			if match:
				latex_link = re.sub(r'<a[^>]*>([^<]+)</a>', converter.convert(match.group(0)).strip(), stripped)
				new_blocks.append(latex_link)
			else:
				new_blocks.append(block)
		return '\n\n'.join(new_blocks)


class Link2Latex:

	def convert(self, instr):
		dom = xml.dom.minidom.parseString(instr)
		link = dom.documentElement
		href = link.getAttribute('href')

		desc = re.search(r'>([^<]+)', instr)
		out = \
                     f"""
			\\href{{{href}}}{{{desc.group(0)[1:]}}}
			"""
		return out


"""
========================= FOOTNOTES =================================

LaTeX footnote support.

Implemented via modification of original markdown approach (place footnote
definition in footnote market <sup> as opposed to putting a reference link).
"""


class FootnoteExtension(markdown.Extension):
	DEF_RE = re.compile(r"(\ ?\ ?\ ?)\[\^([^\]]*)\]:\s*(.*)")
	SHORT_USE_RE = re.compile(r"\[\^([^\]]*)\]", re.M)  # [^a]

	def __init__(self, configs=None):
		self.reset()

	def extendMarkdown(self, md, md_globals):
		self.md = md

		# Stateless extensions do not need to be registered
		md.registerExtension(self)

		# Insert a preprocessor before ReferencePreprocessor
		#index = md.preprocessors.index(md_globals['REFERENCE_PREPROCESSOR'])
		#preprocessor = FootnotePreprocessor(self)
		#preprocessor.md = md
		#md.preprocessors.insert(index, preprocessor)
		md.preprocessors.add('footnotes', FootnotePreprocessor(self), '_begin')

		## Insert an inline pattern before ImageReferencePattern
		FOOTNOTE_RE = r"\[\^([^\]]*)\]"  # blah blah [^1] blah
		#index = md.inlinePatterns.index(md_globals['IMAGE_REFERENCE_PATTERN'])
		#md.inlinePatterns.insert(index, FootnotePattern(FOOTNOTE_RE, self))
		md.inlinePatterns.add('footnotes', FootnotePattern(FOOTNOTE_RE, self), '_begin')

	def reset(self):
		self.used_footnotes = {}
		self.footnotes = {}

	def setFootnote(self, id, text):
		self.footnotes[id] = text


class FootnotePreprocessor:

	def __init__(self, footnotes):
		self.footnotes = footnotes

	def run(self, lines):
		self.blockGuru = BlockGuru()
		lines = self._handleFootnoteDefinitions(lines)

		# Make a hash of all footnote marks in the text so that we
		# know in what order they are supposed to appear.  (This
		# function call doesn't really substitute anything - it's just
		# a way to get a callback for each occurence.

		text = "\n".join(lines)
		self.footnotes.SHORT_USE_RE.sub(self.recordFootnoteUse, text)

		return text.split("\n")

	def recordFootnoteUse(self, match):
		id = match.group(1)
		id = id.strip()
		nextNum = len(list(self.footnotes.used_footnotes.keys())) + 1
		self.footnotes.used_footnotes[id] = nextNum

	def _handleFootnoteDefinitions(self, lines):
		"""Recursively finds all footnote definitions in the lines.

			@param lines: a list of lines of text
			@returns: a string representing the text with footnote
					  definitions removed """

		i, id, footnote = self._findFootnoteDefinition(lines)

		if id:

			plain = lines[:i]

			detabbed, theRest = self.blockGuru.detectTabbed(lines[i + 1:])

			self.footnotes.setFootnote(id, footnote + "\n" + "\n".join(detabbed))

			more_plain = self._handleFootnoteDefinitions(theRest)
			return plain + [""] + more_plain

		else:
			return lines

	def _findFootnoteDefinition(self, lines):
		"""Finds the first line of a footnote definition.

			@param lines: a list of lines of text
			@returns: the index of the line containing a footnote definition.
		"""

		counter = 0
		for line in lines:
			m = self.footnotes.DEF_RE.match(line)
			if m:
				return counter, m.group(2), m.group(3)
			counter += 1
		return counter, None, None


class FootnotePattern(markdown.inlinepatterns.Pattern):

	def __init__(self, pattern, footnotes):
		markdown.inlinepatterns.Pattern.__init__(self, pattern)
		self.footnotes = footnotes

	def handleMatch(self, m, doc):
		sup = doc.createElement('sup')
		id = m.group(2)
		# stick the footnote text in the sup
		self.footnotes.md._processSection(sup, self.footnotes.footnotes[id].split("\n"))
		return sup


def template(template_fo, latex_to_insert):
	tmpl = template_fo.read()
	tmpl = tmpl.replace('INSERT-TEXT-HERE', latex_to_insert)
	return tmpl


# title_items = [ '\\title', '\\end{abstract}', '\\thanks', '\\author' ]
# has_title_stuff = False
# for it in title_items:
#    has_title_stuff = has_title_stuff or (it in tmpl)


def main():

	# stdlib
	import optparse
	usage = \
           """usage: %prog [options] <in-file-path>

		Given a file path, process it using markdown2latex and print the result on
		stdout.

		If using template option template should place text INSERT-TEXT-HERE in the
		template where text should be inserted.
		"""
	parser = optparse.OptionParser(usage)
	parser.add_option(
			'-t', '--template', dest='template', default='', help='path to latex template file (optional)'
			)
	(options, args) = parser.parse_args()
	if not len(args) > 0:
		parser.print_help()
		sys.exit(1)
	inpath = args[0]
	infile = file(inpath)

	md = markdown.Markdown()
	mkdn2latex = LaTeXExtension()
	mkdn2latex.extendMarkdown(md, markdown.__dict__)
	out = md.convert(infile.read())

	if options.template:
		tmpl_fo = file(options.template)
		out = template(tmpl_fo, out)

	print(out)


if __name__ == '__main__':
	main()



#  !/usr/bin/env python
#
#  markdown.py
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
# stdlib
import os
import pathlib
import re
from typing import Union

# 3rd party
import markdown  # type: ignore

# this package
import py2latex.mdx_latex

# this package
from .sectioning import make_chapter

md = markdown.Markdown()
latex_mdx = py2latex.mdx_latex.LaTeXExtension()
latex_mdx.extendMarkdown(md, markdown.__dict__)


def gls(name):
	return rf"\gls{{{name}}}"


def load_markdown(filename: Union[str, pathlib.Path, os.PathLike]) -> str:

	if not isinstance(filename, pathlib.Path):
		filename = pathlib.Path(filename)

	return parse_markdown(filename.read_text())


def parse_markdown(string):
	out = md.convert(string)

	out = re.sub(r"</?root>", '', out)

	out = re.sub(r"gls{([^}]*)}", r"\\gls{\1}", out)
	out = re.sub(r"citep{([^}]*)}", r"~\\citep{\1}", out)
	out = re.sub(r"cite{([^}]*)}", r"~\\cite{\1}", out)

	# text_md = re.sub(r"\*\*(.+)\*\*", r"\\textbf{\1}", text_md)
	# text_md = re.sub(r"\*(.+)\*", r"\\textit{\1}", text_md)
	out = re.sub(r"<sup>(.+)</sup>", r"\\textsuperscript{\1}", out)
	out = re.sub(r"<sub>(.+)</sub>", r"\\textsubscript{\1}", out)
	out = re.sub(r"<sub>(.+)</sub>", r"\\textsubscript{\1}", out)

	return out


def re_escape(string: str) -> str:
	"""
	Escape literal backslashes for use with re.

	:param string:
	:type string:

	:return:
	:rtype:
	"""

	return string.replace("\\", "\\\\")



#  !/usr/bin/env python
#
#  core.py
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from typing import Optional


def make_caption(caption: str) -> str:
	r"""
	Akin to \caption{}

	:param caption:
	:type caption:

	:return:
	:rtype:
	"""

	return rf"\caption{{{caption}}}"


def make_label(label: str) -> str:
	r"""
	Akin to \label{}

	:param label:
	:type label:

	:return:
	:rtype:
	"""

	return rf"\label{{{label}}}"


def begin(environment: str, options: Optional[str] = None) -> str:
	r"""
	Akin to \begin{environment}

	:param environment:
	:type environment:
	:param options:
	:type options:

	:return:
	:rtype:
	"""

	if options:
		return rf"\begin{{{environment}}}{{{options}}}"
	else:
		return rf"\begin{{{environment}}}"


def end(environment: str) -> str:
	r"""
	Akin to \end{environment}

	:param environment:
	:type environment:

	:return:
	:rtype:
	"""

	return rf"\end{{{environment}}}"


def re_escape(string: str) -> str:
	"""
	Escape literal backslashes for use with re.

	:param string:
	:type string:

	:return:
	:rtype:
	"""

	return string.replace("\\", "\\\\")



# stdlib
import os
import pathlib
import re
from typing import Dict, Union

# 3rd party
import yaml

# this package
from py2latex.markdown_parser import parse_markdown


def load_glossary(glossary_file: Union[str, pathlib.Path, os.PathLike]) -> Dict[str, Dict[str, str]]:
	if not isinstance(glossary_file, pathlib.Path):
		glossary_file = pathlib.Path(glossary_file)

	# todo: Validate

	with glossary_file.open() as fp:
		glossary = yaml.load(fp, Loader=yaml.FullLoader)

	for entry, data in glossary["acronyms"].items():
		data["name"] = parse_markdown(data["name"]).strip()
		data["text"] = parse_markdown(data["text"]).strip()

	for entry, data in glossary["glossary"].items():
		data["name"] = parse_markdown(data["name"]).strip()
		data["text"] = parse_markdown(data["text"]).strip()
		data["description"] = parse_markdown(data["description"]).strip()

	return glossary


def escape_prefix(prefix: str) -> str:
	prefix = prefix.replace(" ", "\\ ")
	return prefix


def make_glossary(glossary_dict):
	out = ''

	for acronym, data in glossary_dict["acronyms"].items():
		out += r"\newacronym"

		prefixes = []

		if "prefixfirst" in data:
			prefixes.append(f"prefixfirst={{{escape_prefix(data['prefixfirst'])}}}")
		if "prefix" in data:
			prefixes.append(f"prefix={{{escape_prefix(data['prefix'])}}}")

		if prefixes:
			out += f"[{', '.join(prefixes)}]"

		out += f"{{{acronym}}}{{{data['name']}}}{{{data['text']}}}\n"

	for item, data in glossary_dict["glossary"].items():
		out += fr"""
\newglossaryentry{{{item}}}
{{
name={{{data['name']}}},
text={{{data['text']}}},
description={{{data['description']}}},
}}"""
	return out


def glossary_from_file(filename):
	return make_glossary(load_glossary(filename))



#  This file is managed by `git_helper`. Don't edit it directly
#  Copyright (C) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This file is distributed under the same license terms as the program it came with.
#  There will probably be a file called LICEN[S/C]E in the same directory as this file.
#
#  In any case, this program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# This script based on https://github.com/rocky/python-uncompyle6/blob/master/__pkginfo__.py
#

# stdlib
import pathlib

__all__ = [
		"__copyright__",
		"__version__",
		"modname",
		"pypi_name",
		"py_modules",
		"entry_points",
		"__license__",
		"__author__",
		"short_desc",
		"author",
		"author_email",
		"github_username",
		"web",
		"github_url",
		"project_urls",
		"repo_root",
		"long_description",
		"install_requires",
		"extras_require",
		"classifiers",
		"keywords",
		"import_name",
		]

__copyright__ = """
2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
"""

__version__ = "0.0.4"

modname = "py2latex"
pypi_name = "py2latex"
import_name = "py2latex"
py_modules = []
entry_points = {
		"console_scripts": []
		}

__license__ = "GNU Lesser General Public License v3 or later (LGPLv3+)"

short_desc = "Create LaTeX documents with Python, Markdown and Jinja2."

__author__ = author = "Dominic Davis-Foster"
author_email = "dominic@davis-foster.co.uk"
github_username = "domdfcoding"
web = github_url = f"https://github.com/domdfcoding/py2latex"
project_urls = {
		"Documentation": f"https://py2latex.readthedocs.io",
		"Issue Tracker": f"{github_url}/issues",
		"Source Code": github_url,
		}

repo_root = pathlib.Path(__file__).parent

# Get info from files; set: long_description
long_description = (repo_root / "README.rst").read_text(encoding="utf-8").replace("0.0.4", __version__) + '\n'

install_requires = (repo_root / "requirements.txt").read_text(encoding="utf-8").split('\n')
extras_require = {'all': []}

classifiers = [
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Developers',
		'Operating System :: OS Independent',
		'Topic :: Utilities',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: Implementation :: CPython',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3 :: Only',
		'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',

		]

keywords = ""



#  !/usr/bin/env python
#
#  formatting.py
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
# stdlib
from typing import Union


def bold(val: Union[str, float]) -> str:
	r"""Make the given value bold

	Akin to \textbf{string}

	:param val: The value to make bold

	:return: The formatted string.
	"""

	return fr"\textbf{{{val}}}"


def underline(val: Union[str, float]) -> str:
	r"""Underline the given value

	Akin to \underline{string}

	:param val: The value to underline

	:return: The formatted string.
	"""

	return fr"\underline{{{val}}}"


def italic(val: Union[str, float]) -> str:
	r"""Make the given value italic

	Akin to \textit{string}

	:param val: The value to make italic

	:return: The formatted string.
	"""

	return fr"\textit{{{val}}}"


def latex_subscript(val: Union[str, float]) -> str:
	"""
	Returns the LaTeX subscript of the given value.

	:param val: The value to superscript

	:rtype: str
	"""

	return f'$_{{{val}}}$'


def latex_superscript(val: Union[str, float]) -> str:
	"""
	Returns the LaTeX superscript of the given value.

	:param val: The value to subscript

	:rtype: str
	"""

	return f'$^{{{val}}}$'
